program AddSeabed

! Read in an reflectivity model and all a seabed reflector to it
! mode = 1 : horizontal
! mode = 2 : slanted
! mode = 3 : sinosoidal

  use sep
  use waveprop_mod
  use smooth_mod
  use type_mod

implicit none


 type(velo) :: vel

 logical    :: smooth
 integer    :: ix, iz,  nx, nz, width, nw, mode
 real       :: nr, bound
 real       :: slope, period, pi, upperbound, A
 integer, dimension(:), allocatable :: zloc
 real, dimension(:), allocatable  ::  tempV
 real, dimension(:,:), allocatable :: m, m2

 call sep_init()

   call from_param('smooth', smooth, .false.);
   call from_param('bound',bound, 100.)
   call from_param('width', width, 3)
   call from_param('nw', nw, 0)
   call from_param('mode',mode,1);
   call from_param('slope',slope,0.02);
   call from_param('period', period,200.);
   call from_param('Amp',A,5.);

   if (width .lt. 1) then 
      call erexit('Width must be greater than 1')
   end if


  ! Read original reflectivity model param
   call from_history("n1", vel%z%n);  call from_history("d1", vel%z%d);  call from_history("o1",vel%z%o);
   call from_history("n2", vel%x%n);  call from_history("d2", vel%x%d);  call from_history("o2",vel%x%o);


   pi = 3.14159
   nz = vel%z%n;
   nx = vel%x%n;

   write(0,*) '****************'
   write(0,*) 'smooth, width', smooth, width
   write(0,*) 'dx, dz', vel%x%d, vel%z%d
   write(0,*) 'mode', mode
   write(0,*) 'slope, period, Amp', slope, period, A
   write(0,*) '****************'
   
   allocate( tempV(nz))
   allocate( m(nz,nx), m2(nz+nw,nx))
   allocate(zloc(nx))

   ! Read in the background velocity
    call sep_read( m(1:nz,1:nx))



   ! take 1st derivative of receiver wavefield     
   do ix=1,nx
      tempV(:) = m(:,ix)
      call FirstDer(.true.,tempV, tempV, nz, vel%z%d)
      m(:,ix) = abs(tempV(:))
   end do
   
  
   
   ! Define the location of the seabed
   if (mode .eq. 1) then
   	zloc(:) = nw
   else if (mode .eq. 2) then
   	do ix=1,nx
   	   zloc(ix) =    nw + slope*(nx-ix)
   	end do
   else if (mode .eq. 3) then
   	   do ix=1,nx
           zloc(ix) = nw+A + A*sin(2*pi*(nx/2-ix)/period)
           end do
   end if
   


  ! copy m into m2
   do ix=1,nx
      do iz=1,nz+nw
         if (iz .eq. zloc(ix)) then
         	m2(iz,ix) = bound+1
         else if (iz .gt. zloc(ix)) then
         	m2(iz,ix) = m(iz-nw,ix)
         end if
      end do
  end do


   m2(nz+nw-2:nz+nw,:) = 0.


   do ix=1,nx
      do iz=1,nz+nw
         if (iz .eq. zloc(ix)) then
           m2(iz,ix)=1.5
         else if (m2(iz,ix) .gt. bound) then
          m2(iz,ix) = 1.
         else
          m2(iz,ix)=0.
         end if
      end do
   end do 
     
   ! Smoothing velocity model
    if (smooth) then
    write(0,*) 'Inside the smoother', nz,nx
      call triangle2(width, width, nz+nw, nx, m2, m2)
    end if



   
!----------------------------------------------
! Output

  call sep_write(m2,dim1=nz, dim2=nx, esize=4)
  write(0,*) "vel%z%n, nw", vel%z%n, nw
  call to_history("n1",vel%z%n+nw); call to_history("d1",vel%z%d);  call to_history("o1",vel%z%o);
  call to_history("n2",vel%x%n); call to_history("d2",vel%x%d);  call to_history("o2",vel%x%o);


end program
