program DPtest
  ! perform Dot Product test on RTM oper
  use rtm_operDB
  use waveprop_modDB
  use type_modDB
  use sep
  use smooth_modDB

  implicit none
  logical         :: verb, T, F, smooth, wfldout
  integer          :: nd, n_mod, stat, iy, i, seed
  integer         :: nnz, nnx,  nw
  real, dimension (2) :: dot1 = (/ 0., 0. /), dot2 = (/ 0., 0. /)
  real*8, dimension(:), allocatable :: x, y, Fx, Fty
  real, dimension(:), allocatable :: tmpmodel, tmpdata, tmpfs

  type(velo) :: vel
  type(sou_grid)  :: sou 
  type(rec_grid)  :: rec
  type(bc_values) :: bc

  integer    :: nshot, ix, iz
  real*8       :: rho, pi
  real*8, dimension(:), allocatable   :: fs
  real,   dimension(:,:), allocatable :: tmpvp, m, mm
  real*8, dimension(:,:), allocatable  ::vp
  character(len=128)  ::  sou_, rec_, mod_, velout_, velin_, refl_

  T = .true.
  F = .false. 
  pi = 3.14159265


  nshot = 1

  verb = T
  smooth = T
  wfldout = F
  vel%z%n = 150
  vel%z%d = 10.
  vel%z%o = 0.
  vel%x%n = 1000.
  vel%x%d = 10.
  vel%x%o = 0.
  vel%zw  = 500.

  sou%x%o = 5000.
  sou%x%d = 0.
  sou%x%n = 1
  sou%t%n = 2048 
  sou%t%d = 1.e-3
  sou%t%o = 0.
  sou%zs = 0.
  
!  rec%x%n = vel%x%n
!  rec%x%d = vel%x%d
  rec%x%n = 15
  rec%x%d = 400
  rec%x%o = 100.
  rec%t%n = 2500
  rec%t%d = 1.e-3
  rec%t%o = 0.
  rec%z%n = 1
  rec%z%d = 0.
  rec%z%o = 0.

  rho = 1025.0
  
  bc%Op = 1
  bc%xtapco = 0.0002
  bc%ztapco = 0.0002
  bc%nop = 2
  bc%bndz = 100
  bc%bndx = 20
  bc%ashot = .false.
  bc%shotdir = 1 
  
 

  ! ############ Modify Here ############ 
  rec%t%n=1000
  bc%Op = 1
  ! #####################################
  
  sou_='Sdp1.H'
  rec_='Data/Forcom3.H'
  mod_='Mdp.H'
  velout_='velout.H'
  refl_  = 'refl1.H'
  velin_ = 'Input/veltest1.H'

   ! Read in the source wavefield
   if (sou%t%n .gt. rec%t%n) sou%t%n = rec%t%n

  nnz = vel%z%n
  nnx = vel%x%n
  vel%x%n = vel%x%n + 2*bc%bndx;
  vel%z%n = vel%z%n + 2*bc%bndz;
  nw = (vel%zw-vel%z%o)/vel%z%d + 0.5
  sou%x%n = nshot


!*********************************
   write(0,*) '**************************'
   write(0,*) 'vel: nz, dz, oz ', vel%z%n, vel%z%d, vel%z%o
   write(0,*) 'vel: nx, dx, ox',  vel%x%n, vel%x%d, vel%x%o
   write(0,*) 'vel: zw, nw'         ,  vel%zw, nw
   write(0,*) 'sou: nx, dx, ox',  sou%x%n, sou%x%d, sou%x%o  
   write(0,*)  'sou: nt, dt',     sou%t%n, sou%t%d
   write(0,*) 'rec: nx, dx, ox', rec%x%n, rec%x%d, rec%x%o
   write(0,*) 'nt, dt',          rec%t%n, rec%t%d
   write(0,*) 'nshot, nrz', nshot, rec%z%n
   write(0,*) 'bcOp, verb', bc%Op, verb
   write(0,*) 'smooth, wfldout', smooth, wfldout
   write(0,*) 'orz, drz', rec%z%o, rec%z%d
   write(0,*) 'bndz, bndx', bc%bndz, bc%bndx
   write(0,*) 'xtapco, ztapco', bc%xtapco, bc%ztapco
   write(0,*) '**************************'


  


  n_mod = (nnz-nw)*nnx
  nd = nshot*rec%x%n*rec%t%n*2

! Dot product part (after all the initializaton)
  allocate(x(n_mod), y(nd), Fx(nd), Fty(n_mod));
  allocate(fs(rec%t%n), vp(vel%z%n,vel%x%n))

  allocate(m(nnz,nnx), mm(nnz-nw,nnx))

  allocate(tmpmodel(n_mod), tmpdata(nd))
  allocate(tmpvp(vel%z%n,vel%x%n))
  allocate(tmpfs(rec%t%n))


write(0,*) 'PAP 1'

  call sep_init()


  ! Read in background velocity
!   call sreed(velin_, tmpvp((1+bc%bndz):(vel%z%n-bc%bndz),(1+bc%bndx):(vel%x%n-bc%bndx)),vel%z%n*vel%x%n*4)
 

  write(0,*) 'PAP 2'

  vp(1:vel%z%n/2,:) = 2000.;
  vp(vel%z%n/2+1:vel%z%n,:) = 4000.;
  tmpvp = 1500.;

!    vp is the double version of tmpvp
 do ix=1,vel%x%n
    do iz=1,vel%z%n
       vp(iz,ix) = tmpvp(iz,ix)
    end do
 end do


write(0,*) 'pass 1'




!   call sep_read( fs(1:sou%t%n), "Input/sou1.H")
   call sreed("Input/sou25.H",tmpfs,sou%t%n*4)


write(0,*) 'pass 2'


   do i=1,rec%t%n
      fs(i) = tmpfs(i)
   end do


! Read in reflectivity
!! The bottom 10 lines can be solved with sreed_window
m=0.; mm=0.;


! call sep_read(m(1:nnz,1:nnx),refl_, dim1=nnz,dim2=nnx,esize=4)
! call sreed("refl1.H", m, nnz*nnx*4)

  do ix=1,nnx
    do iz=1,nnz-nw
        mm(iz,ix) = m(iz+nw,ix)
   enddo
 end do

  tmpmodel = RESHAPE(mm,(/(nnz-nw)*nnx/))

! Read in the rec wavefields
!   call sep_read(tmpdata, rec_)



!  x is convert to double from tmpmodel

  do i=1,n_mod
    x(i) = tmpmodel(i)
  end do


!  y = convert to double from tmpdata
  do i=1,nd
    y(i) = tmpdata(i)
  end do
   

  ! Initialization
  stat = L_init(nshot, sou, rec, vel, fs, vp, rho, bc, wfldsou_in=sou_, wfldrec_in=rec_ , wfldmod_in=mod_ , verb_in=T, jsnap_in=99999, seed_in=seed, velout_in=velout_)



  seed = 342342
  do i=1,n_mod
     x(i) = ran(seed)*10.
  end do

  
seed = 75345
  do i=1,nd
     y(i) = ran(seed)*10.
  end do

  
 


  stat = L_oper(F,F, x, Fx)
  stat = L_oper(T,F, Fty, y)
 
!write(0,*) 'Fx'
!do i=1,nd
!write(0,*) i, Fx(i)
!end do


write(0,*) "SUM x, Fx, y, Fty"
write(0,*) "SUM",sum(x),sum(Fx),sum(y),sum(Fty)
write(0,*) size(x), size(Fx), size(y), size(Fty)

  dot1(1) = dot_product(x,Fty)
  dot1(2) = dot_product(y,Fx)


! Second dot test, with Add = True

if ( 1 == 2) then

  seed = 243342
  do i=1,n_mod
     x(i) = ran(seed)*10.
  end do

  seed = 15553  
  do i=1,nd
     y(i) = ran(seed)*100.
  end do

  stat = L_oper(F,F, x, Fx)
  stat = L_oper(T,F, Fty, y)

 

write(0,*) "SUM",sum(x),sum(Fx),sum(y),sum(Fty)

  dot2(1) = dot_product(x,Fty)
  dot2(2) = dot_product(y,Fx)

end if

  write(0,*) 'nt=', rec%t%n, 'bc =', bc%Op
  write(0,*) 'DPtest RTM op'
  write (0,*) dot1
  write (0,*) dot2

  stat = L_close()
  deallocate(x, y, Fx, Fty)
end program DPtest
