program ForwardTWO

  use sep
  use rtm_operTWO
  use waveprop_mod
  use type_mod
  use smooth_mod


  implicit none

  type(velo) :: vel
  type(sou_grid)  :: sougrid
  type(rec_grid)  :: recgrid
  type(bc_values) :: bc

  
  logical                   :: smooth, verb, wfldout
  integer                   :: nshot, ishot, it, id, stat, shotnum
  integer                   :: iz, ix, nz, nx, nt, jsnap
  integer                   :: ii, bndx, bndz, nnz, nnx, nw, nmz, nmx
  character(len=128)        :: rec_, refl_, velout_, wfldsou_, wfldrec_, wfldmod_, out_, rmov_
  character(len=128)        :: vp_
  real                      :: xs, xr
  real                      :: rho
  real, dimension(:), allocatable  :: fs1, tmpdata, tmpmodel

 


  call sep_init()


  
  
  !!======================================================
  !! Parameter reading

   call from_param("verb",verb,.false.);  call from_param("smooth",smooth,.false.);
   call from_param("wfldout", bc%wfldout,.false.);  call from_param("residout",bc%residout,.false.);
   call from_param("wfmout",bc%wfmout,.false.);

   ! Read velin, source and rec file and depth
   call from_param("velin", vp_);
   call from_param("wfldsou", wfldsou_);     call from_param("reflfile",refl_)

   ! Read back ground velocity param
   call from_aux(vp_, "n1", vel%z%n);  call from_aux( vp_, "d1", vel%z%d);  call from_aux( vp_, "o1",vel%z%o);
   call from_aux( vp_, "n2", vel%x%n);  call from_aux( vp_, "d2", vel%x%d);  call from_aux( vp_, "o2",vel%x%o);
   call from_param('zw', vel%zw);
   nw = (vel%zw-vel%z%o)/vel%z%d + 0.5
   call from_aux(refl_,"n1", nmz);  call from_aux(refl_,"n2",nmx);
   

    


   !All shots has the same source wavelet
   call from_aux(wfldsou_, "n3", sougrid%t%n);  call from_aux(wfldsou_,"d3", sougrid%t%d);   call from_aux(wfldsou_,"o3", sougrid%t%o);
   call from_aux(wfldsou_, "n5", nshot);    call from_aux(wfldsou_,"d5", sougrid%x%d);   call from_aux(wfldsou_,"o5", sougrid%x%o);
   call from_param('shotnum',shotnum,0);
   
   call from_param('nrx', recgrid%x%n);    call from_param('drx',recgrid%x%d );     call from_param('orx', recgrid%x%o);
   call from_param('nt', recgrid%t%n);     call from_param('dt', recgrid%t%d);      call from_param('ot', recgrid%t%o);
   call from_param('zr', recgrid%z%o);      
   
   

   ! Read in Density
   call from_param('rho', rho);

   
   !! taper coefficients and padding
   call from_param("bcOp",bc%Op,1);
   call from_param("xtapco",bc%xtapco,0.0002);  call from_param("ztapco",bc%ztapco,0.0002)
   bc%nop = 2
   call from_param("bndz",bc%bndz,bc%nop); if(bc%bndz<bc%nop) bc%bndz=bc%nop; 
   call from_param("bndx",bc%bndx,bc%nop); if(bc%bndx<bc%nop) bc%bndx=bc%nop;


   !! Auxillary output files
   call from_param("outtag", out_);
   call from_param("rmov",rmov_, 'rmov.H')
   call from_param("velout", velout_, "velfile.H");
   call from_param("wfldrec", wfldrec_, "wfldrec.H");
   call from_param("wfldmod", wfldmod_, "wfldmod.H");
   call from_Param("jsnap", jsnap, 100);

   if ((nshot .eq. 1) .and. (sougrid%x%o .eq. vel%x%o)) then
      call from_param('shotnum',shotnum,1)
    	if ((shotnum .gt. 0) .and. (shotnum .le. vel%x%n)) then
    	   sougrid%x%o = sougrid%x%o + (shotnum-1)*sougrid%x%d
      	else
      	   call erexit('shotnum value is invalid. must be >0 and <velgrid%x%n')
      	end if
   end if




   ! Error check

   if ((bc%Op .gt. 2) .or. (bc%Op .lt. 1) ) call erexit('Boundary Cond must be either 1 or 2')
   if (jsnap .le. 0) jsnap=50
   
 !  if (recgrid%x%d .ne. vel%x%d) call erexit('rec dx does not match vel dx')
   
   if (sougrid%t%n .gt. recgrid%t%n) sougrid%t%n = recgrid%t%n

   if (vel%z%n .ne. (nw+nmz)) call erexit('vel nz not equal to model nz + nw')

   
   if (vel%x%n .ne. (nmx)) call erexit('vel nx not equal to model nx')  
   
   if (recgrid%t%n .gt. 5000) call erexit('too many time steps, nt < 5000')
   
 !  if (sougrid%t%d .ne. recgrid%t%d) call erexit('rec dt does not match sou dt')

   if ((recgrid%z%n .eq. 2) .and. (recgrid%z%d .eq. 0)) call erexit('Expect over/under but found dz=0')
   

   !! Definition
   sougrid%x%n = nshot
   nnz = vel%z%n
   nnx = vel%x%n
   bndx = bc%bndx
   bndz = bc%bndz
   vel%z%n = vel%z%n + 2*bndz; 
   vel%x%n = vel%x%n+2*bndx
   nz = vel%z%n;
   nx = vel%x%n;
   nt=recgrid%t%n
   nw = (vel%zw-vel%z%o)/vel%z%d + 0.5


  
   !*********************************
   write(0,*) '**************************'
   write(0,*) 'vel: nz, dz, oz ', nz, vel%z%d, vel%z%o
   write(0,*) 'vel: nx, dx, ox',  nx, vel%x%d, vel%x%o
   write(0,*) 'vel: zw, nw'         ,  vel%zw, nw
   write(0,*) 'sou: nx, dx, ox',  sougrid%x%n, sougrid%x%d, sougrid%x%o  
   write(0,*)  'sou: nt, dt',     sougrid%t%n, sougrid%t%d
   write(0,*) 'rec: nx, dx, ox', recgrid%x%n, recgrid%x%d, recgrid%x%o
   write(0,*) 'nt, dt',          recgrid%t%n, recgrid%t%d
   write(0,*) 'nshot, nrz', nshot, recgrid%z%n
   write(0,*) 'bcOp, verb', bc%Op, verb
   write(0,*) 'smooth, wfldout, residout', smooth, bc%wfldout, bc%residout
   write(0,*) 'orz, drz', recgrid%z%o, recgrid%z%d
   write(0,*) 'bndz, bndx, nop', bc%bndz, bc%bndx, bc%nop
   write(0,*) 'xtapco, ztapco', bc%xtapco, bc%ztapco
   write(0,*) '**************************'


   !----------------------------------------------------------------

   ! Output file
   
  
   call auxputch("n1",'i', recgrid%x%n, trim(out_));     call auxputch("label1",'s','x', trim(out_));
   call auxputch("d1",'f', recgrid%x%d, trim(out_));     call auxputch("o1",'f',recgrid%x%o, trim(out_)); 

   call auxputch("n2",'i',  nt, trim(out_));             call auxputch("label2",'s',  't', trim(out_));
   call auxputch("d2",'f', 0.001, trim(out_));           call auxputch("o2",'f', 0., trim(out_));

!   call auxputch("n3",'i',  2, trim(out_));              call auxputch("label3",'s', 'UD', trim(out_)); 
!   call auxputch("d3",'f', 1., trim(out_));               call auxputch("o3",'f',0., trim(out_));
   
   

   
   ! Allocation
   write(0,*) '*** allocating tmpdata ***', 2*nshot*recgrid%x%n*nt
   allocate(tmpdata(nshot*recgrid%x%n*nt))
      write(0,*) '*** allocating tmpmodel ***', nnx*(nnz-nw)
   allocate(tmpmodel(nnx*(nnz-nw)))




   ! initialize
   tmpdata=0.;
   tmpmodel=0.;

   write(0,*) 'before reading refl model', refl_
   


   ! Read in reflectivity model        
    call sep_read(tmpmodel, refl_);
!   call sep_read(m(1:(nnz-nw),1:nnx),refl_, dim1=(nnz-nw),dim2=nnx,esize=4)
!   tmpmodel = RESHAPE(m,(/(nnz-nw)*nnx/))

write(0,*) 'pass 3'

   if (wfldout) then
      
!      if (exist_file("wfldrec")) then
         call auxputch("n1",'i',nz,trim(wfldrec_));   
         call auxputch("n2",'i',nx,trim(wfldrec_)); 
  !       call auxputch("n1",'i',nz-2*bndz,trim(wfldrec_));   
         call auxputch("d1",'f',vel%z%d,trim(wfldrec_));   call auxputch("o1",'f',vel%z%o,trim(wfldrec_));

  !       call auxputch("n2",'i',nx-2*bndx,trim(wfldrec_));   
         call auxputch("d2",'f',vel%x%d,trim(wfldrec_));   call auxputch("o2",'f',vel%x%o,trim(wfldrec_));

         call auxputch("n3",'i',nt/jsnap,trim(wfldrec_));   
         call auxputch("d3",'f',-1*recgrid%t%d*jsnap,trim(wfldrec_));   call auxputch("o3",'f',recgrid%t%o+(nt-jsnap)*recgrid%t%d,trim(wfldrec_));

!      end if

      if (exist_file("wfldmod")) then

         call auxputch("n1",'i',nz,trim(wfldmod_));   
         call auxputch("d1",'f',vel%z%d,trim(wfldmod_));   call auxputch("o1",'f',vel%z%o,trim(wfldmod_));

         call auxputch("n2",'i',nx,trim(wfldmod_));   
         call auxputch("d2",'f',vel%x%d,trim(wfldmod_));   call auxputch("o2",'f',vel%x%o,trim(wfldmod_));

         call auxputch("n3",'i',nt/jsnap,trim(wfldmod_));   
         call auxputch("d3",'f',recgrid%t%d*jsnap,trim(wfldmod_));   call auxputch("o3",'f',recgrid%t%o,trim(wfldmod_));

      end if

   end if


! write(0,*) ' size of the output', size(tmpdata), recgrid%x%n, nt
! call erexit('Exithere')

   ! Call L_initiate
   stat=L_init(nshot, sougrid, recgrid, vel, wfldsou_, vp_,rho, bc, rmov_in=rmov_, wfldrec_in=wfldrec_, verb_in=verb,jsnap_in=jsnap, velout_in=velout_)



   !! Forward Model: reflectivity ----> data 
   !! adj = false
   stat = L_oper(.false.,.false.,tmpmodel, tmpdata)

   call sep_write(tmpdata, out_)


if (bc%residout) then

     call auxputch("n1",'i',recgrid%x%n,rmov_); 
     call auxputch("d1", 'f', recgrid%x%d, rmov_); call auxputch("o1", 'f', recgrid%x%o, rmov_);
     call auxputch("n2",'i',nt,rmov_); 
     call auxputch("d2", 'f', recgrid%t%d, rmov_); call auxputch("o2", 'f', recgrid%t%o, rmov_);    
!     call auxputch("n3",'i',2,rmov_);    
!     call auxputch("n4",'i',niter,rmov_);    

     call auxputch("label1", 's', 'rx', rmov_); 
     call auxputch("label2", 's', 'nt', rmov_);
!     call auxputch("label3", 's', 'UD', rmov_); 
!     call auxputch("label4", 's', 'niter', rmov_);
 end if

     call sep_close()
     
     deallocate(tmpdata, tmpmodel);

end program
