program GenRefl

! Take derivative of velocity model to generate reflectivity

  use sep
  use waveprop_mod
  use smooth_mod
  use type_mod

implicit none


 type(velo) :: vel

 logical    :: smooth
 integer    :: ix, iz,  nx, nz, width, nw
 real       :: nr, bound
 real, dimension(:), allocatable  ::  tempV, tempV2
 real, dimension(:,:), allocatable :: vp

 call sep_init()

   call from_param('smooth', smooth, .false.);
   call from_param('bound',bound, 100.)
   call from_param('width', width, 3)
   call from_param('nw', nw, 0)

   if (width .lt. 1) then 
      call erexit('Width must be greater than 1')
   end if


  ! Read back ground velocity param
   call from_history("n1", vel%z%n);  call from_history("d1", vel%z%d);  call from_history("o1",vel%z%o);
   call from_history("n2", vel%x%n);  call from_history("d2", vel%x%d);  call from_history("o2",vel%x%o);


   vel%z%n = vel%z%n; 
   vel%x%n = vel%x%n;
   nz = vel%z%n;
   nx = vel%x%n;

   write(0,*) '****************'
   write(0,*) 'smooth, width', smooth, width
   write(0,*) 'dx, dz', vel%x%d, vel%z%d
   write(0,*) '****************'
   
   allocate( vp(nz,nx), tempV(nz))
   allocate( tempV2(nx))


   ! Read in the background velocity
    call sep_read( vp(1:nz,1:nx))




if (1 == 1) then

   ! take 1st derivative of receiver wavefield     
   do ix=1,nx
      tempV(:) = vp(:,ix)
      call FirstDer(.true.,tempV, tempV, nz, vel%z%d)
      vp(:,ix) = abs(tempV(:))
   end do
   
!   do iz=1,nz
!        tempV2(:) = vp(iz,:)
!      call FirstDer(.true.,tempV2, tempV2, nx, vel%x%d)
!      vp(iz,:) = tempV2(:)
!   end do

   vp(nz-2:nz,:) = 0.
   if (nw .ne. 0)   vp(nw-4:nw+4,:) = 0.;


   do ix=1,nx
      do iz=1,nz
         if (vp(iz,ix) .gt. bound) then
          vp(iz,ix)=1.
         else
          vp(iz,ix)=0.
         end if
      end do
   end do 
     
   ! Smoothing velocity model
    if (smooth) then
    write(0,*) 'Inside the smoother', nz,nx
      call triangle2(width, width, nz, nx, vp, vp)
    end if

end if


   
!----------------------------------------------
! Output

  call sep_write(vp,dim1=nz, dim2=nx, esize=4)
  
  call to_history("n1",vel%z%n); call to_history("d1",vel%z%d);  call to_history("o1",vel%z%o);
  call to_history("n2",vel%x%n); call to_history("d2",vel%x%d);  call to_history("o2",vel%x%o);


end program
