program SouGen
! Pre-compute source wavefield with specification defined in a parameter file
! Source wavefield is saved in 1 file using srite_window

  use sep
  use waveprop_mod
  use type_mod
  use smooth_mod
  


  implicit none
  
  

  type(sou_grid)       :: sougrid
  type(rec_grid)       :: recgrid
  type(velo)           :: vel_grid
  type(bc_values)     :: bc

  character(len=128)   :: wfldsou1_, wfldsou2_, sou_
    
    
  logical                              :: verb, wfldout, smooth, deri
  integer                              :: bcOp, nsave, xsnap, istore, shotnum
  integer                              :: nshot, nx, nz, ix, iz, nrx, nt, seed, nw, nnx, nnz
  integer                              ::  izs1, izs2, izr, ixr1, ixr2, irx, ixs, ishot, it
  integer                              :: nop, bndx, bndz, jsnap, ndata, itercount
  real                                 :: dx, ox, dz, oz, dsx, osx, orx, dt, ot
  real                                 ::  zr, orz, drz, drx, rho, maxv
  real                                 :: c0, c1, c2, idx, idz, idt2, nr
  real                                 :: xtapco, ztapco, t
  real*8                               :: time1, time2
  real, dimension(:), allocatable      :: bndz0, bndz1, bndz2, bndx0, bndx1, xtaper, ztaper, fs, tempfr
  real, dimension(:), allocatable      :: xtaper2, ztaper2
  real, dimension(:,:), allocatable    :: vs1sq, vs2sq, vp2, tmpmodel, ctaper,  ctaper2
  real, dimension(:,:), allocatable    :: uo1, um1, up1, uo2, um2, up2, us1Der, us2Der
  real, dimension(:,:,:), allocatable  :: Ran1, Ran2, fr1, fr2
  real, dimension(:,:,:), allocatable  :: us1, us2
  character(len=128)                   :: out_

 !! For OMP
  integer, external  :: omp_get_num_procs, omp_get_thread_num
  integer, external  :: srite_window, sreed_window, getch
  integer            :: inode1, inode, nodes

    call sep_init()

!!======================================================
  !! Parameter reading

   call from_param("verb",verb,.false.);   call from_param("bcOp",bc%Op,1);  
   call from_param("smooth",smooth,.false.); call from_param("wfldout", wfldout,.false.);  

   ! Read back ground velocity param
   call from_history("n1", vel_grid%z%n);  call from_history("d1", vel_grid%z%d);  call from_history("o1",vel_grid%z%o);
   call from_history("n2", vel_grid%x%n);  call from_history("d2", vel_grid%x%d);  call from_history("o2",vel_grid%x%o);

   ! Read source and rec file and depth
   call from_param("soufile", sou_);
   call from_param("outtag", out_);
   

   ! Source and receiver arrangement
   ! all shots has the same source wavelet
   call from_param("osx", sougrid%x%o);      call from_param("dsx", sougrid%x%d); 
   call from_aux(sou_, "n1", sougrid%t%n);  call from_aux(sou_,"d1", sougrid%t%d);   call from_aux(sou_,"o1", sougrid%t%o);
   call from_param('nshot',nshot)
   call from_param('nrx', recgrid%x%n);    call from_param('drx',recgrid%x%d );     call from_param('orx', recgrid%x%o);
   call from_param('nt', recgrid%t%n);     call from_param('dt', recgrid%t%d);      call from_param('ot', recgrid%t%o);
   call from_param('zr', recgrid%z%o);      
   call from_param('zw', vel_grid%zw);
   
   if ((nshot .eq. 1) .and. (sougrid%x%o .eq. vel_grid%x%o)) then
      call from_param('shotnum',shotnum,1)
      		if ((shotnum .gt. 0) .and. (shotnum .le. vel_grid%x%n)) then
       		   sougrid%x%o = sougrid%x%o + (shotnum-1)*sougrid%x%d
      		else
      		   call erexit('shotnum value is invalid. must be >0 and <velgrid%x%n')
      		end if
   end if
   

   ! Read in Density
   call from_param('rho', rho);

   
   !! taper coefficients and padding
   call from_param("xtapco",bc%xtapco,0.0002);  call from_param("ztapco",bc%ztapco,0.0002)
   bc%nop = 2
   call from_param("bndz",bc%bndz,bc%nop); if(bc%bndz<bc%nop) bc%bndz=bc%nop; 
   call from_param("bndx",bc%bndx,bc%nop); if(bc%bndx<bc%nop) bc%bndx=bc%nop;


   !! Auxillary output files
   call from_param("wfldsou", wfldsou1_, "wfldsou.H");
   call from_Param("jsnap", jsnap, 100);



    nx = vel_grid%x%n
    dx = vel_grid%x%d
    ox = vel_grid%x%o
    nz = vel_grid%z%n
    
    dz = vel_grid%z%d
    oz = vel_grid%z%o


    dsx = sougrid%x%d
    osx = sougrid%x%o
    nrx = recgrid%x%n
    orx = recgrid%x%o
    drx = recgrid%x%d
    zr = recgrid%z%o

    nt = recgrid%t%n
    dt = recgrid%t%d
    ot = recgrid%t%o

    nop = bc%nop
    bcOp = bc%Op
    bndx = bc%bndx
    bndz = bc%bndz  
    xtapco = bc%xtapco
    ztapco = bc%ztapco
    nnx = nx 
    nnz = nz 
    nz = nnz + 2*bndz
    nx = nnx + 2*bndx

    nw = (vel_grid%zw - oz)/dz + .5

    c0 = -30./12.
    c1 = +16./12.
    c2 = - 1./12.
    
    ! Parallel Programming
    nodes = omp_get_num_procs()-1;
    call omp_set_num_threads(nodes);
    inode1=0


! ********************
    xsnap = 5
    wfldsou2_ = 'SFor3b.H'
 
! ********************


 !*********************************
   write(0,*) '**************************'
   write(0,*) 'vel: nz, dz, oz ', nz, vel_grid%z%d, vel_grid%z%o
   write(0,*) 'vel: nx, dx, ox',  nx, vel_grid%x%d, vel_grid%x%o
   write(0,*) 'vel: zw, nw'         ,  vel_grid%zw, nw
   write(0,*) 'sou: nx, dx, ox',  sougrid%x%n, sougrid%x%d, sougrid%x%o  
   write(0,*)  'sou: nt, dt',     sougrid%t%n, sougrid%t%d
   write(0,*) 'rec: nx, dx, ox', recgrid%x%n, recgrid%x%d, recgrid%x%o
   write(0,*) 'nt, dt',          recgrid%t%n, recgrid%t%d
   write(0,*) 'nshot, nrz', nshot, recgrid%z%n
   write(0,*) 'bcOp, verb', bc%Op, verb
   write(0,*) 'smooth, wfldout', smooth, wfldout
   write(0,*) 'orz, drz', recgrid%z%o, recgrid%z%d
   write(0,*) 'bndz, bndx, nop', bc%bndz, bc%bndx, bc%nop
   write(0,*) 'xtapco, ztapco', bc%xtapco, bc%ztapco
   write(0,*) 'nodes', nodes
   write(0,*) '**************************'
 !***********************************


    allocate(fs(nt), tempfr(nt), vp2(nz,nx)) 
    allocate(fr1(nshot,nrx,nt), fr2(nshot,nrx,nt))
    allocate(vs1sq(nz-nw,nx), vs2sq(nz+nw,nx))
    
    ! Initializing
    fs=0.; tempfr=0.; vp2=0.;
    fr1=0.; fr2=0.;
    vs1sq=0.;  vs2sq=0.;
    

    ! Take 1st derivative of source function 
     write(0,*) 'Der of Sou wavefield'
     call FirstDer(.true., fs ,fs ,nt,dt)
    

       
    izs1 = bndz
    izs2 = bndz
  
 
    !----------------------------------------------------------------
    !! one-way absorbing BC and Tapersetup  
    allocate(bndz0(nx),bndz1(nx))      
    allocate(bndx0(nz),bndx1(nz))
    allocate(ztaper(bndz))
    allocate(xtaper(bndx))
    allocate(ctaper(bndz,bndx))
    allocate(ztaper2(bndz))
    allocate(xtaper2(bndx))
    allocate(ctaper2(bndz,bndx))
    if (bcOp .eq. 2) allocate(Ran1(nz,2*bndx,nshot), Ran2(2*bndz,nx-2*bndx,nshot))



    !----------------------------------------------------------------
   ! Read in the source wavefields
   call sep_read( fs(1:sougrid%t%n), sou_)
   fs(sougrid%t%n:nt)=0.;

   ! Read in the background velocity
   call sep_read( vp2(1+bndz:nz-bndz,1+bndx:nx-bndx))
   do iz=1,bndz; vp2(   iz  ,:) = vp2( 1+bndz,:); vp2(nz-iz+1,:) = vp2(nz-bndz,:); end do !padding
   do ix=1,bndx; vp2(:,   ix  ) = vp2(:, 1+bndx); vp2(:,nx-ix+1) = vp2(:,nx-bndx); end do !padding

   
    idx = 1./dx**2; idz = 1./dz**2; idt2= 1./dt**2
    
    
     call  abcsetup(nx,nz,vp2,dt,dx, dz, nop, bndz0,bndz1,bndx0,bndx1)      
     call  tapersetupCos(bndz,bndx,ztaper,xtaper,ctaper)
     call  tapersetupCos2(bndz,bndx,ztaper2,xtaper2,ctaper2)

write(0,*) 'nx, nz, nw', nz, nx,nw

    do ix=1,nx
       do iz=1,nz      
          vp2(iz,ix) = vp2(iz,ix)*vp2(iz,ix)
          vs2sq(iz+nw,ix) = vp2(iz,ix)    
       end do

       do iz=1,nw
          vs2sq(iz+bndz,ix) = vp2(iz+bndz,ix)
       end do

       do iz=1,nz-nw
          vs1sq(iz,ix) = vp2(iz+nw,ix)
       end do

    end do

write(0,*) 'pass vs2sq'

     do iz=1,bndz
	 vs1sq(   iz  ,:) = vs1sq( 1+bndz,:); vs1sq(nz-nw-iz+1,:) = vs1sq(nz-nw-bndz,:); 
	 vs2sq(   iz  ,:) = vs2sq( 1+bndz,:); vs2sq(nz+nw-iz+1,:) = vs2sq(nz+nw-bndz,:); 
	 vp2(   iz  ,:) = vp2( 1+bndz,:); vp2(nz-iz+1,:) = vp2(nz-bndz,:); 
     end do 


     do ix=1,bndx
        vs1sq(:,   ix  ) = vs1sq(:, 1+bndx); vs1sq(:,nx-ix+1) = vs1sq(:,nx-bndx); 
        vs2sq(:,   ix  ) = vs2sq(:, 1+bndx); vs2sq(:,nx-ix+1) = vs2sq(:,nx-bndx); 
        vp2(:,   ix  ) = vp2(:, 1+bndx); vp2(:,nx-ix+1) = vp2(:,nx-bndx); 
     end do


     call taperingCos(nz-nw,nx,bndz,bndx,ztaper2,xtaper2,ctaper2, vs1sq)
     call taperingCos(nz+nw,nx,bndz,bndx,ztaper2,xtaper2,ctaper2, vs2sq)


     ! Smoothing velocity model
      call triangle2(5, 5, nz-nw, nx, vs1sq, vs1sq)
      call triangle2(5, 5, nz-nw, nx, vs1sq, vs1sq)
      call triangle2(5, 5, nz-nw, nx, vs1sq, vs1sq)
      call triangle2(5, 5, nz-nw, nx, vs1sq, vs1sq)
      call triangle2(5, 5, nz-nw, nx, vs1sq, vs1sq)
      call triangle2(5, 5, nz-nw, nx, vs1sq, vs1sq)
      call triangle2(5, 5, nz+nw, nx, vs2sq, vs2sq)
      call triangle2(5, 5, nz+nw, nx, vs2sq, vs2sq)
      call triangle2(5, 5, nz+nw, nx, vs2sq, vs2sq)
      call triangle2(5, 5, nz+nw, nx, vs2sq, vs2sq)
      call triangle2(5, 5, nz+nw, nx, vs2sq, vs2sq)
      call triangle2(5, 5, nz+nw, nx, vs2sq, vs2sq)




   if (bcOp .eq. 2) then
        seed = 1000
        maxv = 3000.
        do ishot=1,nshot
  	  call rbcsetup(ishot, vp2, Ran1, Ran2, bndx, bndz, nx, nz, seed, maxv)
        end do
   end if
  

    nsave = (nt)/xsnap
    write(0,*) 'allocating us', (nnz-nw)*nnx*nsave*nshot
    allocate( us1(nnz-nw,nnx,nsave), us2(nnz-nw,nnx,nsave) )

    write(0,*) 'allocating 6X', nz*nx
    allocate(up1(nz-nw,nx), um1(nz-nw,nx), uo1(nz-nw,nx))
    allocate(up2(nz+nw,nx), um2(nz+nw,nx), uo2(nz+nw,nx))
    allocate(us1Der(nz-nw,nx), us2Der(nz+nw,nx))
     


    ! Initialization
    tmpmodel=0
    !!! data_init should be commented out
  !!  call data_init(nz,nx,nsave,nshot, ssave_)




    ! -------------------------------------------------------------
    ! Pre-compute source wavefield

    ! Start timer
    call cpu_time(time1)


  
    do ishot=1,nshot
       
       ixs = bndx + 1 + ((osx+(ishot-1)*dsx) - ox)/dx

       ! initializing the array
       !$OMP parallel do ordered schedule(dynamic,1)
       do ix=1,nx
          do iz=1,nz-nw
             up1(iz,ix)=0.; um1(iz,ix)=0.; uo1(iz,ix)=0.;
             up2(iz,ix)=0.; um2(iz,ix)=0.; uo2(iz,ix)=0.;
          end do          
          do iz=nz-nw+1,nz+nw
             up2(iz,ix)=0.;  um2(iz,ix)=0.; uo2(iz,ix)=0.;
          end do
       end do
       !$OMP end parallel do
       

       do it=1,nt

           !! 4th order laplacian
	   if (bcOp .eq. 1) then
	      if (nop .eq. 1) then
                 call laplac2(.true., um1, uo1,up1, vs1sq, dt, nx, nz-nw, idx, idz, nop)
                 call laplac2(.true., um2, uo2,up2, vs2sq, dt, nx, nz+nw, idx, idz, nop)
              else if (nop .eq. 2) then
                 call laplac4(.true., um1, uo1,up1, vs1sq, dt, nx, nz-nw, idx, idz, nop)
                 call laplac4(.true., um2, uo2,up2, vs2sq, dt, nx, nz+nw, idx, idz, nop)
              end if   
           else
!              call laplac4RAN(ishot,um,uo,up,vp2,Ran1,Ran2,dt,nx,nz, bndx, bndz, idx,idz,2)
           end if
           
!           vel_grid%z%n = nz-nw
!           call absorbingBCHale(.true. , vel_grid,dt, vp2, um1,uo1,up1)
!	   vel_grid%z%n = nz


	  ! Take second derivative (in time) of source wavefield
	  if (it .gt. 1) then
		  do ix=1,nx
		     do iz=1,nz-nw
		        us1Der(iz,ix) = (up1(iz,ix) - 2*uo1(iz,ix) + um1(iz,ix))
       		        us2Der(iz,ix) = (up2(iz,ix) - 2*uo2(iz,ix) + um2(iz,ix))
		     end do
		     
		     do iz=nz-nw+1,nz+nw
		     	us2Der(iz,ix) = (up2(iz,ix) - 2*uo2(iz,ix) + um2(iz,ix))
		     end do
		     
	  	  end do
	  else
	  	  do ix=1,nx
		     do iz=1,nz-nw
		        us1Der(iz,ix) = 0.
       		        us2Der(iz,ix) = 0.
		     end do
		     do iz=nz-nw+1,nz+nw
		     	us2Der(iz,ix) = 0.
		     end do
	  	  end do   
	  end if


           !! time step
           !$OMP parallel do ordered schedule(dynamic,1)
           do ix=1,nx
              do iz=1,nz-nw
                 um1(iz,ix)=uo1(iz,ix); 
                 uo1(iz,ix)=up1(iz,ix);
                 um2(iz,ix)=uo2(iz,ix); 
                 uo2(iz,ix)=up2(iz,ix);
              end do          
              do iz=nz-nw+1,nz+nw
                 um2(iz,ix)=uo2(iz,ix); 
                 uo2(iz,ix)=up2(iz,ix);
              end do
           end do
          !$OMP end parallel do
          


           ! Injecting source/rec wavelet into the finite diff grid
           t = (it-1)*dt
           if (t .lt. 0.1) then
            uo1(izs1,ixs) =  uo1(izs1,ixs) + fs(it)  
	    uo2(izs2,ixs) = uo2(izs2,ixs) + fs(it)
           end if
           
           !! If absorbing boundary (bcOp .eq. 1) apply the operation here
           !! Else, let wavefield be scattered by random boundary (bcOp .eq. 1)
           if (bcOp .eq. 1) then
              !! taper
		
	       call taperingCos(nz-nw,nx,bndz,bndx,ztaper,xtaper,ctaper, uo1)
	       call taperingCos(nz-nw,nx,bndz,bndx,ztaper,xtaper,ctaper, up1)
	       call taperingCos(nz-nw,nx,bndz,bndx,ztaper,xtaper,ctaper, um1)
	       call taperingCos(nz+nw,nx,bndz,bndx,ztaper,xtaper,ctaper, uo2)
	       call taperingCos(nz+nw,nx,bndz,bndx,ztaper,xtaper,ctaper, up2)
	       call taperingCos(nz+nw,nx,bndz,bndx,ztaper,xtaper,ctaper, um2)

             !! Absorbing boundary
             call absorbingBC1(nz-nw,nx,nop,bndz0,bndz1,bndx0,bndx1,uo1,um1)
             call absorbingBC1(nz+nw,nx,nop,bndz0,bndz1,bndx0,bndx1,uo2,um2)             
          end if




           ! Store source wavefield to memory for every xsnap time step
           if ((mod(it,xsnap) == 0) .or. (xsnap .eq. 1) ) then
              istore = it/xsnap
              do ix=1,nnx
                 do iz=1,nnz-nw
		    us1(iz,ix,istore) = us1Der(iz+bndz,ix+bndx)
		    us2(iz,ix,istore) = us2Der(iz+bndz+2*nw,ix+bndx)		    
!	            us1(iz,ix,istore) = uo1(iz+bndz,ix+bndx)
!	            us2(iz,ix,istore) = uo2(iz+bndz+2*nw,ix+bndx)
                 end do
              end do	     
           end if

 
           if (verb .and. (mod(it,50) .eq. 1)) then
              write(0,*) 'Sou: it, ishot, sum', it, ishot, sum(uo1), sum(uo2)
           end if
        
 
        	
             if(mod(it,jsnap)==0 .and. wfldout .and. ishot==1 ) then
!                  call sep_write(uo1,wfldsou1_)
                  call sep_write(uo2,wfldsou2_)
             end if
           


       end do !(end source time loop)
       
           call sep_write(us1 , trim(out_), dim1=nnz-nw, dim2=nnx, dim3=nsave,esize=4)
           call sep_write(us2 , trim(out_), dim1=nnz-nw, dim2=nnx, dim3=nsave,esize=4)
       
     end do !(end shot loop)

  

! ------------------------------------------------------------
! ------ Output file -----------------------------------------

 
 
 
! call sep_write(us1, dim1=nnz-nw, dim2=nnx, dim3=nsave, dim4=nshot, esize=4)

! call sep_write(us2, dim1=nnz-nw, dim2=nnx, dim3=nsave, dim4=nshot, esize=4)

write(0,*) 'nz, nnz, nw', nz, nnz, nw

   call auxputch("n1",'i',nnz-nw, trim(out_)); call auxputch("d1",'f', dz, trim(out_));  call auxputch("o1",'f', oz, trim(out_));
   call auxputch("n2",'i', nnx, trim(out_)); call auxputch("d2",'f', dx, trim(out_));  call auxputch("o2",'f', ox, trim(out_));
   call auxputch("n3",'i', nsave, trim(out_)); call auxputch("d3",'f', dt*xsnap, trim(out_));  call auxputch("o3",'f', ((xsnap-1)*dt+ot), trim(out_));
   call auxputch("n4",'i', 2, trim(out_)); call auxputch("d4",'f', 1, trim(out_));  call auxputch("o4",'f', 0, trim(out_));        
   call auxputch("n5",'i', nshot, trim(out_)); call auxputch("d5",'f', dsx, trim(out_));  call auxputch("o5",'f', osx, trim(out_));       
   call auxputch("label1", 's', 'nz', trim(out_));
   call auxputch("label2", 's', 'nx', trim(out_));
   call auxputch("label3", 's', 'nt', trim(out_));
   call auxputch("label4", 's', 'UD', trim(out_));
   call auxputch("label5", 's', 'nshot', trim(out_));



write(0,*) 'wfldsou1', wfldsou1_

 if ((wfldout) .and. (1 == 2)) then

 	   call auxputch("n1",'i',nz-nw,trim(wfldsou1_));
 	   write(0,*) 'P 1'
   	   call auxputch("d1",'f',dz,trim(wfldsou1_));   
   	   write(0,*) 'P 2'
           call auxputch("n2",'i',nx,trim(wfldsou1_));  
           write(0,*) 'P 3'
   	   call auxputch("d2",'f',dx,trim(wfldsou1_));   
   	   write(0,*) 'P 4'
   	   call auxputch("label1", 's', 'Z', trim(wfldsou1_));
   	   write(0,*) 'P 5'
     	   call auxputch("label2", 's', 'X', trim(wfldsou1_));
     	   write(0,*) 'P 6'
           call auxputch("n3",'i',nt/jsnap,trim(wfldsou1_));
           write(0,*) 'P 7'
           call auxputch("d3",'f', dt*jsnap, trim(wfldsou1_));   
           write(0,*) 'P 8'
  	   call auxputch("label3", 's', 'time', trim(wfldsou1_));
           write(0,*) 'P 9'
  end if
  if ((wfldout) .and. 1 == 1) then

 	   call auxputch("n1",'i',nz+nw,trim(wfldsou2_));  
   	   call auxputch("d1",'f',dz,trim(wfldsou2_));   
           call auxputch("n2",'i',nx,trim(wfldsou2_));  
   	   call auxputch("d2",'f',dx,trim(wfldsou2_));  
   	   call auxputch("label1", 's', 'Z', trim(wfldsou2_));
     	   call auxputch("label2", 's', 'X', trim(wfldsou2_));
           call auxputch("n3",'i',nt/50,trim(wfldsou2_));   
           call auxputch("d3",'f', dt*jsnap, trim(wfldsou2_));   
  	   call auxputch("label3", 's', 'time', trim(wfldsou2_));

 end if

   ! --------------------------------------------------------------

   deallocate( fs, tempfr, vp2)
   deallocate( fr1, fr2, vs1sq, vs2sq)
   deallocate( bndz0, bndz1, bndx0, bndx1)
   deallocate( ztaper, xtaper, ctaper)
   deallocate( ztaper2, xtaper2, ctaper2)
   deallocate( us1, us2)
   deallocate( up1, um1, uo1, up2, um2, uo2)

   call sep_close()

! Start timer
    call cpu_time(time2)
    write(0,*) 'CPU runtime is:', time2-time1, 'sec'


write(0,*) 'program should exit'

end program
