program modcgstep
! Instead of inputing r and dr. This program takes the value of the relevant dot product
! and calculate alpha and beta
! Input: mfile, dmfile, sfile
! Input: rrdgg, rrdrr (if Forget=T)
! Input: ggdgg, ssdss, rrdrr, ggdss, rrdss,  (if Forget=F)
! Output: m+alpha*dm+beta*s
! Output: alpha,beta

  

  use sep
  
  implicit none
  
  
  logical                               :: Forget
  integer                               :: n1, n2, i1, i2
  integer                               :: n1aux, n2aux
  real                                  :: d1, d2, o1, o2
  real                                  :: ggdgg, ggdss, ssdss, ggdrr, rrdrr, ssdrr
  real                                  :: alpha, beta, determ
  real, dimension(:,:), allocatable     :: m, dm, s, update
  character(len=128)                    :: mfile_, dmfile_, sfile_, outfile_

  
  
  
  call sep_init()
  
  
  call from_param('Forget', Forget, .true.)
  if (.not. Forget) then
      call from_param('ggdgg',ggdgg);
      call from_param('ggdss', ggdss);
      call from_param('ssdss', ssdss);
      call from_param('ssdrr', ssdrr);
      
  else
      call from_param('rrdrr', rrdrr);
  end if
  
  call from_param('ggdrr',ggdrr);
  
  ! obtain filename
  call from_param('mfile',mfile_);
  call from_param('dmfile', dmfile_);  
  call from_param('sfile', sfile_);  
  call from_param('outtag', outfile_);

  
  call from_aux(mfile_,"n1",n1);     call from_aux(mfile_,"d1",d1);       call from_aux(mfile_,"o1",o1);
  call from_aux(mfile_,"n2",n2);     call from_aux(mfile_,"d2",d2);       call from_aux(mfile_,"o2",o2);
  
   call from_aux(dmfile_,"n1",n1aux);   call from_aux(dmfile_,"n2",n2aux); 
  if ((n1aux .ne. n1) .or. (n2aux .ne. n2)) call erexit('size of mfile and dmfile not match!');
  
  allocate(m(n1,n2), dm(n1,n2), s(n1,n2), update(n1,n2))
  
  do i2=1,n2
     do i1=1,n1
    	m(i1,i2)=0.
  	dm(i1,i2)=0.
  	s(i1,i2)=0.
  	update(i1,i2)=0.
     end do
  end do
  
  call sep_read(m,mfile_);
  call sep_read(dm,dmfile_);
  
  if (.not. Forget) call sep_read(s,sfile_);
  
  ! Calculate alpha and beta here
  if (Forget) then      ! steepest decent
     beta = 0.d0
     if (ggdgg .eq. 0) then
         call erexit('modcgstep: grad vanish identically')
     end if
     alpha = -ggdrr/rrdrr
  else
     if ((ggdgg .eq. 0.) .or. (ssdss .eq. 0.)) then
          alpha=0.;
          beta =0.;
     else
     
          determ = ggdgg*ssdss*max(1.d0 - (ggdss/ggdgg)*(ggdss/ssdss), 1.d-12)
          alpha = (ssdss*ggdrr - ggdss*ssdrr)/determ
          beta  = (-1.*ggdss*ggdrr + ggdgg*ssdrr)/determ
     end if
  end if     


  ! Write the new alpha and beta to a file
  OPEN(53,FILE=outfile_, ACCESS='APPEND')  
    write(53,*) alpha
    write(53,*) beta
  CLOSE(53)
  


  ! Update model and the previous step (both s and ss)
  
  do i2=1,n2
     do i1=1,n1     
      update(i1,i2) = alpha*dm(i1,i2) + beta*s(i1,i2)
      m(i1,i2)      = m(i1,i2) + update(i1,i2)     
     end do
  end do     
  
  ! write the new model and the new s to file
  call sep_write(m,mfile_)
  call sep_write(update,sfile_)
  
  
  deallocate(m, s, dm, update)
 
  call sep_close()

end program
