module rtm_operTWODB


  use sep
  use type_modDB
  use waveprop_modDB
  use smooth_modDB

  implicit none
  
  type(velo)       :: vel_grid

  logical, private                             :: verb, wfrout, wfmout, residout, UDout
  integer, private                             :: bcOp, nsave, xsnap, istore
  integer, private                             :: nshot, nx, nz, nrx, nt, seed, nw, nnx, nnz
  integer, private                             :: i, ishot, it, ix, iz
  integer, private                             ::  izs1, izs2, izr, ixr, irx, ixs
  integer, private                             :: nop, bndx, bndz, jsnap, ndata, itercount
  real*8, private                                :: dx, ox, dz, oz, dsx, osx, drx, orx, dt, ot
  real*8, private                                ::  zr, orz, drz, rho, maxv
  real*8, private                                :: c0, c1, c2, idx, idz, idt2, nr
  real*8, private                                :: xtapco, ztapco, delp
  real*8, private                              :: time1, time2
  integer, dimension(:), allocatable, private  :: tapebd
  real*8, dimension(:), allocatable,private      :: bndz0, bndz1, bndz2, bndx0, bndx1, xtaper, ztaper, fs, tmpdata
  real*8, dimension(:), allocatable, private     :: xtaper2, ztaper2
  real*8, dimension(:,:), allocatable,private    :: vp2, tmpmodel, ctaper, ctaper2, Wup
  real*8, dimension(:,:), allocatable,private    :: uo, um, up
  real*8, dimension(:,:,:), allocatable, private :: fr1
  real*8, dimension(:,:,:), allocatable,private  :: us
  real, dimension(:,:,:), allocatable,private  :: ussingle
  character(len=128), private                  ::  wfldrec_, wfldmod_, ssave_, rmov_
  
  
   !! For OMP
  integer, external  :: omp_get_num_procs, omp_get_thread_num
  integer, external  :: srite_window, sreed_window, getch
  integer            :: inode1, inode, nodes, Totalnodes
  
  contains 

  integer function L_init(nshot_in, sougrid_in, recgrid_in, velgrid_in, us_in, vp_in, rho_in, bc_in, &
  		  rmov_in, wfldrec_in, wfldmod_in, jsnap_in, verb_in, velout_in) result(stat)


    optional ::  rmov_in, wfldrec_in, wfldmod_in, jsnap_in, verb_in

    type(sou_grid)       :: sougrid_in
    type(rec_grid)       :: recgrid_in
    type(velo)           :: velgrid_in
    type(bc_values)     :: bc_in

    logical              :: verb_in
    integer              :: nshot_in, ix, iz, jsnap_in, seed_in
    real*8                 :: rho_in
    character(len=128)   :: wfldrec_in, wfldmod_in, velout_in, rmov_in, us_in, vp_in

    itercount = 0
    nshot = nshot_in

    nx = velgrid_in%x%n
    dx = velgrid_in%x%d
    ox = velgrid_in%x%o
    nz = velgrid_in%z%n
    dz = velgrid_in%z%d
    oz = velgrid_in%z%o


    dsx = sougrid_in%x%d
    osx = sougrid_in%x%o
    nrx = recgrid_in%x%n
    orx = recgrid_in%x%o
    drx = recgrid_in%x%d
    zr = recgrid_in%z%o

    nt = recgrid_in%t%n
    dt = recgrid_in%t%d
    ot = recgrid_in%t%o


    bcOp = bc_in%Op
    bndx = bc_in%bndx
    bndz = bc_in%bndz
    nnx = nx - 2*bndx
    nnz = nz - 2*bndz
    nw = (velgrid_in%zw - oz)/dz + .5
    nop = bc_in%nop
    residout = bc_in%residout
    UDout    = bc_in%UDout
    wfrout  = bc_in%wfrout
    wfmout   = bc_in%wfmout

    xsnap = sougrid_in%t%d/recgrid_in%t%d + 0.5
    jsnap = jsnap_in
    nsave = sougrid_in%t%n 

    c0 = -30./12.
    c1 = +16./12.
    c2 = - 1./12.
    
    vel_grid = velgrid_in

! ********************
    write(0,*) 'xsnap', xsnap, sougrid_in%t%d, recgrid_in%t%d
    
! ********************
    allocate( vp2(nz,nx)) 
    allocate(fr1(nshot,nrx,nt))
	
write(0,*) 'pass 4'
    ndata = nshot*nrx*nt
    allocate(tmpdata(ndata))
    

    vp2=0.;
    fr1=0.; 
    tmpdata=0.;


write(0,*) 'pass 5'

    if (residout .and. present(rmov_in)) then
       rmov_ = rmov_in
    end if

    if (wfrout) then
       wfldrec_ = wfldrec_in
    end if
    write(0,*) 'wfmout', wfmout
    if (wfmout) then
       wfldmod_ = wfldmod_in
    end if


    if (present(verb_in)) verb = verb_in
       
    izs1 = bndz

  
 
  
   ! Read in the background velocity
!   call sep_read( vp2(1+bndz:nz-bndz,1+bndx:nx-bndx), vp_in, esize=4)

!    vp2 is the velocity squared
 do ix=1+bndx,nx-bndx
    do iz=1+bndz,nz-bndz
       vp2(iz,ix) = 2500.
    end do
 end do


   write(0,*) 'pass 7'
   do iz=1,bndz; vp2(   iz  ,:) = vp2( 1+bndz,:); vp2(nz-iz+1,:) = vp2(nz-bndz,:); end do !padding
   write(0,*) 'pass 8' 
   do ix=1,bndx; vp2(:,   ix  ) = vp2(:, 1+bndx); vp2(:,nx-ix+1) = vp2(:,nx-bndx); end do !padding
write(0,*) 'pass 7'

    write(0,*) 'allocating 6X', nz*nx
    allocate(up(nz,nx))
    allocate(um(nz,nx))
    allocate(uo(nz,nx))
    
    write(0,*) 'nz, nx'
    
    !----------------------------------------------------------------
    !! one-way absorbing BC and Tapersetup  
    allocate(bndz0(nx),bndz1(nx))      
    allocate(bndx0(nz),bndx1(nz))
    allocate(ztaper(bndz))
    allocate(xtaper(bndx))
    allocate(ctaper(bndz,bndx))
    allocate(ztaper2(bndz))
    allocate(xtaper2(bndx))
    allocate(ctaper2(bndz,bndx))


    bndz0 =0.; bndz1=0.; 
    bndx0=0.; bndx1=0.;
    ztaper=0.; xtaper=0.;
    ctaper=0.; ztaper2=0.;
    xtaper2=0.; ctaper2=0.;

    !----------------------------------------------------------------

   
    idx = 1./dx**2; idz = 1./dz**2; idt2= 1./dt**2
    
    call  abcsetup(nx,nz,vp2,dt,dx, dz, nop, bndz0,bndz1,bndx0,bndx1)      
     call  tapersetupCos(bndz,bndx,ztaper,xtaper,ctaper)
     call  tapersetupCos2(bndz,bndx,ztaper2,xtaper2,ctaper2)

   !  waterv=1500.


    do ix=1,nx
       do iz=1,nz      
          vp2(iz,ix) = vp2(iz,ix)*vp2(iz,ix) 
       end do
    end do

write(0,*) 'pass 7'


   call taperingCos(nz,nx,bndz,bndx,ztaper2,xtaper2,ctaper2,vp2)



     ! Smoothing velocity model
      call triangle2(5, 5, nz, nx, vp2, vp2)
      call triangle2(5, 5, nz, nx, vp2, vp2)
      call triangle2(5, 5, nz, nx, vp2, vp2)
      call triangle2(5, 5, nz, nx, vp2, vp2)
      call triangle2(5, 5, nz, nx, vp2, vp2)
      call triangle2(5, 5, nz, nx, vp2, vp2)


  

     



 write(0,*) 'pass 8'
    ! Allocation
    write(0,*) 'allocating us', (nnz-nw)*nnx*nsave
    allocate(us(nnz-nw,nnx,nsave))
    allocate(ussingle(nnz-nw,nnx,nsave))
    us=0.;
    ussingle=0.;

    write(0,*) 'allocating 2X', nz*nx
    allocate(tmpmodel(nnz-nw,nnx))
    allocate(Wup(nnz-nw,nnx),tapebd(nnx))
    
    tapebd = 10.
    
    do ix=1,nnx
    	do iz=1,nnz-nw
		tmpmodel(iz,ix) = 0.;
		Wup(iz,ix) = 1.;
	end do
    end do   	

   write(0,*) 'pass Wup check'
 
! ********
! Turn off Wup 
! ********* 
!   call WuptaperSetup(tapebd,10,Wup,nnz-nw,nnx)



   ! Read in the sou wavefield
   write(0,*) 'before reading sou wavefield', trim(us_in)
   call sep_read( ussingle, us_in, esize=4)

   do i=1,nsave
      do ix=1,nnx
        do iz=1,nnz-nw
           us(iz,ix,i) = ussingle(iz,ix,i)
        end do
      end do
   end do 
     
    uo=0.; um=0.; up=0.;



    ! Initialization
    tmpmodel=0



   ! --------------------------------------------------------------
 
    stat=0

  end function L_init

  !---------------------------------------------------------------------------
  !***************************************************************************
  !---------------------------------------------------------------------------
  integer function L_oper(adj, add, model, data) result(stat)


    logical, intent(in)         :: adj, add
    real*8, dimension(:)  :: model, data 

    
    integer                   :: ix, iz, iop, id, i
    integer                   :: i1,i2,i3,i4
    real*8                      :: datasum, frsum


    stat = 1

    ! Start timer
    call cpu_time(time1)

    
    
    if (.not. add) then
       if (adj) then
          model = 0.
       else
          data=0.
       end if
    end if

    ! Parallel Programming

    Totalnodes = omp_get_num_procs()
    nodes = Totalnodes-2;
    call omp_set_num_threads(nodes);
    inode1=0
    write(0,*) 'OMP Paralleling'
    write(0,*) 'using', nodes, 'out of', Totalnodes
    
    if (adj) then      !!! ### RTM ### !!!



     ! Copy rec wavefield from 1D array back to a 4D array

     
        tmpdata = data(1:ndata)      
	fr1 = RESHAPE(tmpdata, (/nshot,nrx,nt/))



         ! take 1st derivative of receiver wavefield
         if (1 == 2) then
            do irx=1,nrx
              do ishot=1,nshot
                if (irx*ishot .eq. 1) write(0,*) 'take derivative of Rec'
                tmpdata(1:nt) = fr1(ishot,irx,1:nt)
       	        call FirstDer(adj, tmpdata, tmpdata, nt, dt)
                fr1(ishot,irx,1:nt) = tmpdata(1:nt)
                        
              end do
           end do
         end if

      


      !! Initizalize the array
         tmpmodel=0.;
         
         
     ! Shot loop
       do ishot=1,nshot

          
       ! initializing the array
       !$OMP parallel do ordered schedule(dynamic,5) private(ix,iz)
       do ix=1,nx
          do iz=1,nz
             up(iz,ix)=0.; um(iz,ix)=0.; uo(iz,ix)=0.;
          end do          
       end do
       !$OMP end parallel do



          do it=nt,1,-1
         
       
       		
              !! 4th order laplacian
               call laplac4(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop)



 !          call absorbingBCHale(adj, vel_grid,dt, vp2, um,uo,up)

           !! time step
           !$OMP parallel do ordered schedule(dynamic,5) private(ix,iz)
           do ix=1,nx
              do iz=1,nz
                 um(iz,ix)=uo(iz,ix); 
                 uo(iz,ix)=up(iz,ix);
 
              end do          
           end do
          !$OMP end parallel do
          
          

               !! taper
               if (1 == 1) then
                    call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, uo)
                    call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, up)
                    call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, um)
		end if

              !! Absorbing boundary
              call absorbingBC1(nz,nx,nop,bndz0,bndz1,bndx0,bndx1,uo,um)
    


             ! Inject receiver signal  
               izr= (zr-oz)/dz + 1.5 + bndz
               !$OMP parallel do ordered schedule(dynamic,1) private(irx,ixr)
               do irx=1,nrx
                   ixr = (orx+(irx-1)*drx -ox)/dx + 1.5 + bndx
                   uo(izr,ixr) = uo(izr,ixr) + fr1(ishot, irx, it)  
               end do 
               !$OMP end parallel do
               if ((mod(it,100) .eq. 0)) write(0,*) 'adjoint ishot, it', ishot, it





            !!! RTM: IMAGING CONDITION
            ! Multiply the source and receiver wavefield at each time step (zero lag)
 
       
               if ((it .lt. nt) .and. ((xsnap .eq. 1) .or. (mod(it,xsnap) .eq. 0))) then
	        istore = it/xsnap
	        !$OMP parallel do ordered schedule(dynamic,5) private(ix,iz)
                do ix=1,nnx
                   do iz=1,(nnz-nw)
	              tmpmodel(iz,ix) = tmpmodel(iz,ix) + us(iz,ix,istore)*uo(iz+bndz+nw,ix+bndx)

                   enddo
                enddo
                !$OMP end parallel do
             end if
       




        enddo ! (end rec time loop)
     enddo ! (shot loop)
     


do ix=1,nnx
  do iz=1,(nnz-nw)
     tmpmodel(iz,ix) = Wup(iz,ix)*tmpmodel(iz,ix)
  end do
end do



    else  !!! ### FORWARD Modelling ### !!!


      ! Copy reflectivity model from 1D array back to 2D array
       tmpmodel = RESHAPE( model, (/nnz-nw,nnx/))
       
      ! take first derivative of the model
      if (1 == 2) then
         do ix=1,nnx
         	  tmpdata(1:(nnz-nw)) = tmpmodel(:,ix)
        	  call FirstDer(adj, tmpdata, tmpdata, nnz-nw, dz)
        	  tmpmodel(:,ix) = tmpdata(1:(nnz-nw))
         end do
      end if
  
      !! Rec Loop
       fr1=0.; 

      ! Shot loop
      do ishot=1,nshot
        
        ! initializing the array
        !$OMP parallel do ordered schedule(dynamic,5) private(ix,iz)
         do ix=1,nx
            do iz=1,nz
               up(iz,ix)=0.; um(iz,ix)=0.; uo(iz,ix)=0.;
            end do          
         end do
         !$OMP end parallel do
        
           
               do it=1,nt

		  if (1 == 1) then
                    call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, uo)
                    call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, up)
                    call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, um)
		   end if

                   !! 4th order laplacian
                   call laplac4(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop)


                   !! time step
                  !$OMP parallel do ordered schedule(dynamic,5) private(ix,iz)
                   do ix=1,nx
                      do iz=1,nz
                         um(iz,ix)=uo(iz,ix); 
                         uo(iz,ix)=up(iz,ix);
                       end do          
                   end do
                   !$OMP end parallel do
 
                   call absorbingBC1(nz,nx,nop,bndz0,bndz1,bndx0,bndx1,uo,um)


                    ! Forward modelling Injection
                   if ( (xsnap .eq. 1) .or. (mod(it,xsnap) .eq. 0) )then
		      istore = it/xsnap   
		      !$OMP parallel do ordered schedule(dynamic,5) private(ix,iz)
                      do ix=1,nnx
                         do iz=1,(nnz-nw)
                             uo(iz+nw+bndz,ix+bndx) = uo(iz+nw+bndz,ix+bndx) + us(iz,ix,istore)*Wup(iz,ix)*tmpmodel(iz,ix) 
                         enddo
                      enddo
                      !$OMP end parallel do
                   end if


                   !! Collect data
		     izr= (zr-oz)/dz + 1.5 + bndz
		     !$OMP parallel do ordered schedule(dynamic,1) private(irx,ixr)
                     do irx=1,nrx
                        ixr = (orx+(irx-1)*drx -ox)/dx + 1.5 + bndx
                        fr1(ishot, irx, it) = fr1(ishot, irx, it)  + uo(izr,ixr)
                     end do
                     !$OMP end parallel do
                     if ((mod(it,100) .eq. 1)) write(0,*) 'forward sum, it, ishot', it, ishot,sum(fr1)


            
                  if(mod(it,jsnap)==0 .and. wfrout .and. ishot==1 ) then
      	 	      write(0,*) 'output ur', ishot, it, sum(uo)
!                        call sep_write(uo(1:nz,1:nx),wfldrec_,dim1=nz,dim2=nx,esize=4)
                  end if
                  
                  
                  


               end do ! (rec time loop)
        enddo ! (shot loop)

 
        
        
   end if



if (wfrout) then
         call auxputch("n1",'i',nz,trim(wfldrec_));   
         call auxputch("d1",'f',dz,trim(wfldrec_));   call auxputch("o1",'f',oz,trim(wfldrec_));
	 call auxputch("label1", 's', 'Z', trim(wfldrec_));

         call auxputch("n2",'i',nx,trim(wfldrec_));   
         call auxputch("d2",'f',dx,trim(wfldrec_));   call auxputch("o2",'f',ox,trim(wfldrec_));
	 call auxputch("label2", 's', 'Z', trim(wfldrec_));

         call auxputch("n3",'i',nt/jsnap,trim(wfldrec_));   
         call auxputch("d3",'f',-1*dt*jsnap,trim(wfldrec_));   
	 call auxputch("o3",'f',ot+(nt-jsnap)*dt,trim(wfldrec_));
 	 call auxputch("label3", 's', 'time', trim(wfldrec_));
end if

if (wfmout) then 	 
 	 
 	 call auxputch("n1",'i',nnz-nw,trim(wfldmod_));   
         call auxputch("d1",'f',dz,trim(wfldmod_));   call auxputch("o1",'f',oz,trim(wfldmod_));
	 call auxputch("label1", 's', 'Z', trim(wfldmod_));

         call auxputch("n2",'i',nnx,trim(wfldmod_));   
         call auxputch("d2",'f',dx,trim(wfldmod_));   call auxputch("o2",'f',ox,trim(wfldmod_));
	 call auxputch("label2", 's', 'Z', trim(wfldmod_));

         call auxputch("n3",'i',nt/jsnap,trim(wfldmod_));   
         call auxputch("d3",'f',-1*dt*jsnap,trim(wfldmod_));   
	 call auxputch("o3",'f',ot+(nt-jsnap)*dt,trim(wfldmod_));
 	 call auxputch("label3", 's', 'time', trim(wfldmod_));
 	 
end if
   ! ------------------------------------------------------

   
    

    if (.not. adj) then
    
      ! take 1st derivative of receiver wavefield
      if (1 == 2) then
            do irx=1,nrx
              do ishot=1,nshot
                if (irx*ishot .eq. 1) write(0,*) 'take derivative of Rec'
                tmpdata(1:nt) = fr1(ishot,irx,1:nt)
       	        call FirstDer(adj, tmpdata, tmpdata, nt, dt)
                fr1(ishot,irx,1:nt) = tmpdata(1:nt)
              end do
           end do
      end if
    
	ndata = nshot*nt*nrx
       write(0,*) 'cp data back', ndata
       data(1:ndata) = RESHAPE(fr1,SHAPE=(/nshot*nt*nrx/))
    else
    
       ! take first derivative of the model
       if (1 == 2) then
	   do ix=1,nnx
       		  tmpdata(1:(nnz-nw)) = tmpmodel(:,ix)
       		  call FirstDer(adj, tmpdata, tmpdata, nnz-nw, dz)
       		  tmpmodel(:,ix) = tmpdata(1:(nnz-nw))
       	   end do
       end if
    
       model = RESHAPE(tmpmodel,SHAPE=(/ (nnz-nw)*nnx /) )

   
    end if

    

    ! end timer
    call cpu_time(time2)
    write(0,*) 'time for L_oper', time2-time1, 'sec'

    stat =0

    end function

    !---------------------------------------------------------------------------
    !---------------------------------------------------------------------------
    integer function L_close() result(stat)

      stat=1

	write(0,*) 'dalloc 1'
      deallocate(fr1, vp2)

      	write(0,*) 'dalloc 2'
      deallocate(bndz0, bndz1)
	write(0,*) 'dalloc 3'
      deallocate(bndx0, bndx1)
	write(0,*) 'dalloc 4'
      deallocate(xtaper, ztaper, ctaper)
      deallocate(xtaper2, ztaper2, ctaper2)
	write(0,*) 'dalloc 5'
      deallocate(um, up, uo)

	write(0,*) 'dalloc 6'
      deallocate(tmpmodel)

        write(0,*) 'dealloc 7'
      deallocate(us)
      deallocate(tmpdata)
      
      stat=0

    end function L_close

end module
