module step_iter_mod
  implicit none
  real,allocatable,private :: ss(:),s(:),gt(:)
  double precision, private,allocatable :: gd(:)
  integer, allocatable,private :: k(:)

  integer, private :: ieq_1,ieq_2

  real :: scale_solve
  
contains
! ===========================================================
   subroutine iterstep_close()
     deallocate(ss,s,gt,gd,k)
   end subroutine
   

! ===========================================================   
   function iterstep(forget,x,g,rr,gg) result(ieq)
     real :: rr(:),gg(:),x(:),g(:)
     logical :: forget
     double precision :: alpha,beta,gamma
     integer :: ieq, iter
     logical :: found
      if(.not. allocated(s)) then
        allocate(s(size(x)),ss(size(rr)),gt(size(rr)),gd(size(rr)),k(size(rr)))
      end if
     if(forget) then
       ieq_1=-1
       ieq_2=-1
       s=0;
       ss=0;
       call minabs(size(rr),rr,gg,ieq,gamma)
       alpha=gamma; beta=0;
       s=g*alpha+beta*s
       ss=alpha*gg+beta*ss
       x=x+s
       rr=rr+ss
       ieq_2=ieq_1
       ieq_1=ieq
    else
        do iter=1, size(x)
           gt=ss(ieq_1)*gg-gg(ieq_1)*ss
           call minabs(size(rr),rr,gt,ieq,gamma)
           alpha=ss(ieq_1)*gamma
           beta=-gg(ieq_1)*gamma
write(0,*) ieq_1, ieq_2, ieq, gamma
        if(gamma==0. .or. ieq == ieq_2)  then
write(0,*) "out of inner loop"           
return
        end if
        s=g*alpha+beta*s
        ss=alpha*gg+beta*ss
        x=x+s
        rr=rr+ss
        ieq_2=ieq_1
        ieq_1=ieq
        
     end do
  end if
end function iterstep

! ===========================================================   
subroutine minabs( n, f, w, ieq,alpha)
integer i,n,low,large,ml,mh,mlt,mht,itry,j,ieq,i1,i2,it
real gn,gp,t,er,gnt,gpt,gplx,gmix,small,grad,mys,mp,mn ,tst,mymin
double precision :: alpha
          real    w(:),f(:)
small = 0.
do i=1,n
	k(i) = i
end do
low = 1
large = n
ml = n
mh = 1
gn = 0.
gp = 0.
do itry = 1,n 
   j = k( low+mod( (large-low)/3+itry, large-low+1) ) !choose an equation to test
   ieq=j
   if( abs(w(j)) /= 0.)then
      t = f(j)/(w(j))   !we have our test step size
      f(j) = w(j)*t     !set f to w*t to guarantee 0 in loop
      do i = low,large  !loop through all equations in current range
         j = k(i)      !select equation in current set
         er = f(j)-w(j)*t !calculate the error given current step size
         gd(j) = 0.  
         if     ( er>small) then
            gd(j) = -w(j)    !if error is positive scale gg by negative num
            
         else if( er<(-small)) then
            gd(j) = +w(j)    !copy gg into g
            
         else
            gd(j) = 0.
            
            
         end if
      end do
      call split(n,low,large,k,gd,mlt,mht) !divide gg into positive and negative piles
      gnt = gn
      do i = low,mlt
         gnt = gnt+gd(k(i))  !sum up negative wts
      end do
      gpt = gp
      do i = mht,large
         gpt = gpt+gd(k(i)) !sum up positive wts
      end do
      gplx = 0.
      gmix = 0.
      do i = mlt,mht   !loop through small ws?
         j = k(i)
         if (w(j)<0.) then
            gplx = gplx-w(j) 
            gmix = gmix+w(j)
         end if
         if (w(j)>0.) then
            gplx = gplx+w(j)
            gmix = gmix-w(j)
         end if
      end do
      grad = gnt+gpt  !sum of gradients
      if ((grad+gplx)*(grad+gmix)<-0.) then
         exit
      end if
      if (grad>=0.) low = mht+1
      if (grad<=0) large = mlt-1
      if (low>large) then
         exit
      end if
      if (grad>=0.) gn = gnt+gmix
      if (grad<=0.) 	gp = gpt+gplx
      if (grad+gplx==0.) ml = mlt
   end if
   if (grad+gmix==0.) mh = mht
end do
alpha = - t
end subroutine minabs

! ===========================================================
subroutine split(n,low,large,k,g,ml,mh)
  !	given g(k(i)),i=low,large
  !	then rearrange k(i),i=low,large and find ml and mh so that
  !	(g(k(i)),i=low,(ml-1)) < 0. and
  !	(g(k(i)),i=ml,mh)=0. and
  !	(g(k(i)),i=(mh+1),large) > 0.
  integer :: n
  double precision   g(n)
  integer k(n)
  integer low,large,ml,mh,keep,i,ii
  ml = low
  mh = large
  major: do while(1==1)
     ml = ml-1
     do while(ml <n)
        ml = ml+1
        if(g(k(ml))>=0) exit
     end do
     minor: do while(1==1)
        mh = mh+1
        do while(mh >1 )
           mh = mh-1
           if(g(k(mh))<=0.) exit
        end do
        keep = k(mh)
        k(mh) = k(ml)
        k(ml) = keep
        if (g(k(ml))/=g(k(mh))) exit minor		! Break out of enclosing "repeat{"
        do i = ml,mh 
           ii = i
           if (g(k(i))/=0.0) then
              go to 30
           end if
        end do
        exit major		! Break out of two enclosing "repeat{"'s
30      keep = k(mh)
        k(mh) = k(ii)
        k(ii) = keep
     end do minor
  end do major
  return
end  subroutine split
! ===========================================================

end module step_iter_mod
